/* 
 * Proview   $Id: pwr_version.c,v 1.2 2005-09-01 14:57:47 claes Exp $
 * Copyright (C) 2005 SSAB Oxelsund AB.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 2 of 
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with the program, if not, write to the Free Software 
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* pwr_version.c -- create pwr_version.h
   Creates the file pwr_version.h which is used to
   define common version variables.  */

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "pwr.h"
#include "wb_ldh.h"
#include "rt_load.h"
#include "wb_pwrs.h"
#include "wb_pwrb.h"

typedef char pwr_tVersionStr[32];

int main (
  int argc,
  char *argv[]
) {
  time_t      tp;
  struct tm * time_tm;
  char	      timestr[80];
  FILE *      fp;

  pwr_tVersionStr   SmdVersion;
  pwr_tVersionStr   BmdVersion;
  pwr_tVersionStr   WbdbVersion;
  pwr_tVersionStr   LffVersion;
  pwr_tVersionStr   Version;

  /* get arguments */
  if (argc < 2) {
    printf("Error creating pwr_version.h: wrong number of arguments\n");
    exit(0);
  }
  
  strcpy(Version, *++argv);
  strcpy(SmdVersion, pwrs_cSmdVersionStr);
  strcpy(BmdVersion, pwrb_cBmdVersionStr);
  strcpy(WbdbVersion, ldh_cWbdbVersionStr);
  strcpy(LffVersion, load_cLffVersionStr);

  /* get build time */
  time(&tp);
  time_tm = localtime(&tp);
  strftime(timestr, sizeof(timestr), "%d %B %Y", time_tm);

  fp = fopen("pwr_version.h", "w");
  fprintf(fp, "\n");
  fprintf(fp, "#ifndef pwr_version_h\n");
  fprintf(fp, "#define pwr_version_h\n");
  fprintf(fp, "\n");
  fprintf(fp, "/*\n");
  fprintf(fp, "  File: pwr_version.h\n");
  fprintf(fp, "\n");
  fprintf(fp, "  NOTE  !!!\n");
  fprintf(fp, "    This file is generated by the build procedure\n");
  fprintf(fp, "    and should not be edited.\n");
  fprintf(fp, "\n");
  fprintf(fp, "  Description:\n");
  fprintf(fp, "    Defines version varibles to use in messages to user.\n");
  fprintf(fp, "\n");
  fprintf(fp, "*/\n");
  fprintf(fp, "\n");
  fprintf(fp, "#include <time.h>\n");
  fprintf(fp, "\n");
  fprintf(fp, "#ifndef pwr_h\n");
  fprintf(fp, "#include \"pwr.h\"\n");
  fprintf(fp, "#endif\n");
  fprintf(fp, "\n");
  fprintf(fp, "#if defined OS_VMS\n");
  fprintf(fp, "#define pwrv_cOpSys              \"OpenVMS\"\n");
  fprintf(fp, "#elif defined OS_ELN\n");
  fprintf(fp, "#define pwrv_cOpSys              \"VAXELN\"\n");
  fprintf(fp, "#else\n");
  fprintf(fp, "#define pwrv_cOpSys              \"Unknown Operating System\"\n");
  fprintf(fp, "#endif\n");
  fprintf(fp, "\n");
  fprintf(fp, "#if defined VAX || defined __VAX\n");
  fprintf(fp, "#define pwrv_cHardware           \"VAX\"\n");
  fprintf(fp, "#elif defined __Alpha_AXP\n");
  fprintf(fp, "#define pwrv_cHardware           \"Alpha AXP\"\n");
  fprintf(fp, "#else\n");
  fprintf(fp, "#define pwrv_cHardware           \"Unknown Hardware\"\n");
  fprintf(fp, "#endif\n");
  fprintf(fp, "\n");
  fprintf(fp, "#define pwrv_cBuildTime          ((time_t) %d)\n", (int)tp);
  fprintf(fp, "#define pwrv_cBuildTimeStr       \"%s\"\n", timestr);
  fprintf(fp, "\n");
  fprintf(fp, "#define pwrv_cPwrVersion         ((pwr_tVersion)((%c << 24) + (%c << 16) + (%c << 8) + \'%c\'))\n",
    Version[5], Version[3], Version[1], Version[0]);
  fprintf(fp, "#define pwrv_cPwrVersionStr      \"%s\"\n", Version);
  fprintf(fp, "\n");
  fprintf(fp, "#define pwrv_cSmdVersion         ((pwr_tVersion)((%c << 24) + (%c << 16) + (%c << 8) + \'%c\'))\n",
    SmdVersion[5], SmdVersion[3], SmdVersion[1], SmdVersion[0]);
  fprintf(fp, "#define pwrv_cSmdVersionStr      \"%s\"\n", SmdVersion);
  fprintf(fp, "\n");
  fprintf(fp, "#define pwrv_cBmdVersion         ((pwr_tVersion)((%c << 24) + (%c << 16) + (%c << 8) + \'%c\'))\n",
    BmdVersion[5], BmdVersion[3], BmdVersion[1], BmdVersion[0]);
  fprintf(fp, "#define pwrv_cBmdVersionStr      \"%s\"\n", BmdVersion);  
  fprintf(fp, "\n");
  fprintf(fp, "#define pwrv_cWbdbVersion        ((pwr_tVersion)((%c << 24) + (%c << 16) + (%c << 8) + \'%c\'))\n",
    WbdbVersion[5], WbdbVersion[3], WbdbVersion[1], WbdbVersion[0]);
  fprintf(fp, "#define pwrv_cWbdbVersionStr     \"%s\"\n", WbdbVersion);  
  fprintf(fp, "#define pwrv_cWbdbVersionShortStr \"%c%c%c\"\n",
    WbdbVersion[0], WbdbVersion[1], WbdbVersion[3]);
  fprintf(fp, "#define pwrv_cWbdbVersionStr     \"%s\"\n", WbdbVersion);  
  fprintf(fp, "\n");
  fprintf(fp, "#define pwrv_cLffVersion         ((pwr_tVersion)((%c << 24) + (%c << 16) + (%c << 8) + \'%c\'))\n",
    LffVersion[5], LffVersion[3], LffVersion[1], LffVersion[0]);
  fprintf(fp, "#define pwrv_cLffVersionStr      \"%s\"\n", LffVersion);  
  fprintf(fp, "\n");
  fprintf(fp, "#endif\n");
  fclose(fp);

  return 0;
}
