/* 
 * Proview   $Id: rt_rtt_dir.c,v 1.2 2005-09-01 14:57:56 claes Exp $
 * Copyright (C) 2005 SSAB Oxelsund AB.
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 2 of 
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with the program, if not, write to the Free Software 
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* rt_rtt_dir.c
   This module contains routines for handling files and directories in rtt. */


/*_Include files_________________________________________________________*/



#if defined(OS_ELN)
# include stdio
# include stdarg
# include stdlib
# include signal
# include string
# include descrip
# include starlet
# include lib$routines
# include $vaxelnc
# include $kernelmsg
# include $elnmsg
# include $get_message_text
# include $file_utility
#elif defined(OS_VMS)
# include <stdio.h>
# include <stdlib.h>
# include <ctype.h>
# include <string.h>
# include <starlet.h>
# include <rms.h>
# include <descrip.h>
#elif defined(OS_LYNX) || defined(OS_LINUX)
# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include <ctype.h>
# include <string.h>
# include <dirent.h>
#endif

#include "pwr.h"
#include "rt_rtt.h"
#include "rt_rtt_functions.h"
#include "rt_rtt_dir.h"
#include "rt_rtt_msg.h"

/*_Globala variabler______________________________________________________*/

/*_Local proceduer________________________________________________________*/

/* Globala procedurer_____________________________________________________*/

#define ODD(a)	(((int)(a) & 1) != 0)
#define EVEN(a)	(((int)(a) & 1) == 0)


/*************************************************************************
*
* Name:		rtt_search_file
*
* Typ		int
*
* Typ		Parameter	IOGF	Beskrivning
*
* Beskrivning: 
*		Sker efter en fil. 
*		Filnamnet kan innehlla wildcard.
*		Vid skning av flera filer mha wildcard anropas
*		rutinen frsta gngen med new=1 och sedan med new=0.
*		Tv skningar med wildcard kan ej pg parallellt efter
*		som skvillkor lagras i interna variabler.
*
* Parametrar
*	file_name	*char	I  Namn p fil, kan innehlla wildcard.
*	found_file	*char	O  Hittad fil.
*	new		int	I  1 vid ej wildcard skning och vid frsta
*				   wildcardskningsanropet. 0 vid skning
*				   efter fler filer med samma wildcard.
*
**************************************************************************/

int rtt_search_file( 		char 	*name , 
				char	*found_file,
				int	new)
{
#if defined (OS_VMS)
	static struct NAM 	nam1;
	static struct FAB 	fab1;
	static struct	XABDAT	dat1; 
	static char		result_name[NAM$C_MAXRSS];
	static char		expand_name[NAM$C_MAXRSS];
	int 			sts;


	if ( new == RTT_DIR_SEARCH_END)
	  return 1;

	if ( new == RTT_DIR_SEARCH_INIT)
	{
	  /* Initialize RMS user structures for the checking file. */
	  fab1  	   = cc$rms_fab;
	  fab1.fab$l_xab = (char *) &dat1;
	  fab1.fab$l_nam = &nam1;
	  fab1.fab$w_ifi = 0;
  
	  dat1  	   = cc$rms_xabdat;
	  dat1.xab$l_nxt = 0;

	  fab1.fab$l_fna = name;
	  fab1.fab$b_fns = strlen (name);

	  nam1  	   = cc$rms_nam;
	  nam1.nam$l_rsa = result_name;
	  nam1.nam$b_rss = sizeof(result_name);
	  nam1.nam$l_esa = expand_name;
	  nam1.nam$b_ess = sizeof(expand_name);

	  sts = sys$parse(&fab1);
	  if (EVEN(sts)) return sts;
	}

	if ( new == RTT_DIR_SEARCH_INIT || new == RTT_DIR_SEARCH_NEXT)
	{
	  sts= sys$search (&fab1);
	  if (EVEN(sts)) return sts;
	  result_name[nam1.nam$b_rsl] = 0;
	  strcpy( found_file, result_name);
	}
	return 1;
#elif defined (OS_ELN)

	static ELN$DIR_FILE *dir_file;
	VARYING_STRING(255) search_name, volume_name, directory_name,
			file_name, server_name;
	char	directory_str[255], file_str[255],
		volume_str[255], server_str[255];
	int	sts;
	static FILE$ATTRIBUTES_RECORD *fattr;

	if ( new == RTT_DIR_SEARCH_INIT)
	{
	  dir_file = calloc( 1, sizeof(*dir_file));
	  CSTRING_TO_VARYING( name, search_name);
	  eln$directory_open( &dir_file, &search_name, &volume_name, 
		&directory_name, &sts, &server_name, &fattr);
	  if ( EVEN(sts)) return sts;
	}

	if ( new == RTT_DIR_SEARCH_INIT || new == RTT_DIR_SEARCH_NEXT)
	{
	  eln$directory_list( &dir_file, &directory_name, &file_name, &sts,
		&fattr);
	  if (EVEN(sts)) return sts;

	  VARYING_TO_CSTRING( directory_name, directory_str);
	  VARYING_TO_CSTRING( file_name, file_str);
	  VARYING_TO_CSTRING( volume_name, volume_str);
	  VARYING_TO_CSTRING( server_name, server_str);
	  strcpy( found_file, server_str);
	  strcat( found_file, volume_str);
	  strcat( found_file, directory_str);
	  strcat( found_file, file_str);
	}
	if ( new == RTT_DIR_SEARCH_END)
	  eln$directory_close( &dir_file, &sts);

	return sts;
#elif defined (OS_LYNX) || defined(OS_LINUX)

	static DIR *directory;
	static char pattern[80];
	char	dev[2], dir[200], dir2[200], file[80], type[80];
	char	cwd[200];
	int	version;
	int	found;
	int	wildcard;
	struct dirent *dir_entry;

	if ( new == RTT_DIR_SEARCH_INIT)
	{
	  rtt_parse_filename( name, dev, dir, file, type, &version);
	  rtt_replace_env( dir, dir2);
	  if ( strchr( dir2, '/') == 0)
	  {
	    /* Add cwd to the path */
	    if ( getcwd( cwd, sizeof(cwd)) == NULL)
	      return RTT__NOFILE;
	    strcpy( dir, cwd);
	    strcat( dir, dir2);
	  }
	  else
	    strcpy( dir, dir2);

	  directory = opendir( dir);
	  if ( directory == NULL)
	    return RTT__NOFILE;

	  strcpy( pattern, file);
	  strcat( pattern, type);
#if 0
	  rtt_toupper( pattern, pattern);
#endif
	}

	if ( new == RTT_DIR_SEARCH_INIT || new == RTT_DIR_SEARCH_NEXT)
	{
	  if ( strchr( pattern, '*') != 0)
	    wildcard = 1;
	  else
	    wildcard = 0;

	  found = 0;
	  while ( (dir_entry = readdir( directory)) != NULL)
	  {
	    if ( dir_entry->d_name[0] == '.')
	      continue;

	    if ( wildcard)
	    {
	      if ( rtt_wildcard( pattern, dir_entry->d_name) == 0)
	      {
	        strcpy( found_file, dir_entry->d_name);
	        found = 1;
	      }
	    }
	    else
	    {
	      if ( strcmp( pattern, dir_entry->d_name) == 0)
	      {
	        strcpy( found_file, dir_entry->d_name);
	        found = 1;
	      }
	    }
	    if ( found)
	      break;
	  }
	  if ( !found)
	    return RTT__NOFILE;
	  else
	    return RTT__SUCCESS;
	}

	if ( new == RTT_DIR_SEARCH_END)
	  closedir( directory);

	return RTT__SUCCESS;
#else
  	return 0;
#endif
	return RTT__SUCCESS;
}  


/*************************************************************************
*
* Name:		rtt_parse_filename
*
* Typ		int
*
* Typ		Parameter	IOGF	Beskrivning
*
* Beskrivning: 
*	Delar upp ett filnamn i device, directory, namn, typ och version.
*
* Parameterar
*	filenam		*char	I  Filnamn som ska delas.
*	dev		*char	O  device.
*	dir		*char	O  directory.
*	file		*char	O  filnamn.
*	type		*char	O  typ.
*	version		*int	O  version.
*
**************************************************************************/

int	rtt_parse_filename( 	char	*filename,
				char	*dev,
				char	*dir,
				char	*file,
				char	*type,
				int	*version)
{
#if defined (OS_VMS)
	struct NAM 	nam1;
	struct FAB 	fab1;
	struct	XABDAT	dat1; 
	char		result_name[NAM$C_MAXRSS];
	char		expand_name[NAM$C_MAXRSS];
	int 		sts;
	int		nr;


	/* Initialize RMS user structures for the checking file. */
	fab1  	   = cc$rms_fab;
	fab1.fab$l_xab = (char *) &dat1;
	fab1.fab$l_nam = &nam1;
	fab1.fab$w_ifi = 0;
  
	dat1  	   = cc$rms_xabdat;
	dat1.xab$l_nxt = 0;

	fab1.fab$l_fna = filename;
	fab1.fab$b_fns = strlen (filename);

	nam1  	   = cc$rms_nam;
	nam1.nam$l_rsa = result_name;
	nam1.nam$b_rss = sizeof(result_name);
	nam1.nam$l_esa = expand_name;
	nam1.nam$b_ess = sizeof(expand_name);

	sts = sys$parse(&fab1);
	if (EVEN(sts)) return sts;

	strcpy( dev, nam1.nam$l_dev);
	dev[nam1.nam$b_dev] = 0;
	strcpy( dir, nam1.nam$l_dir);
	dir[nam1.nam$b_dir] = 0;
	strcpy( file, nam1.nam$l_name);
	file[nam1.nam$b_name] = 0;
	strcpy( type, nam1.nam$l_type);
	type[nam1.nam$b_type] = 0;
	nam1.nam$l_ver[nam1.nam$b_ver] = 0;
	nr = sscanf( nam1.nam$l_ver+1, "%d", version);
	if ( nr != 1)
	  *version = -1;

	return 1;
#elif defined (OS_ELN)
	char	*s, *t;
	char	ldev[200];
	char	ldir[200];
	char	lfile[200];
	char	ltype[80];
	char	lversion[80];
	int	sts;

	if ( s = strstr( filename, "::"))
	  s += 2;
	else
	  s = filename;
	strcpy( ldev, s);

	/* Device */
	if ( s = strchr( ldev, ':'))
	  s++;
	else
	  s = ldev;
	strcpy( ldir, s);
	*s = 0;

	/* Directory */
	if ( s = strchr( ldir, '>'))
	  s++;
	else
	{
	  if ( s = strchr( ldir, ']'))
	    s++;
	  else
	    s = ldir;
	}
	strcpy( lfile, s);
	*s = 0;

	/* File */
	if ( (s = strchr( lfile, '.')) == 0)
	  s = lfile;
	strcpy( ltype, s);
	*s = 0;

	/* Directory */
	if ( s = strchr( ltype, ';'))
	  t = s++;
	else
	{
	  if ( s = strchr( ltype + 1, '.'))
	    t = s++;
	  else
	  {
	    s = ltype;
	    t = s + strlen(s);
	  }
	}
	strcpy( lversion, t);
	*t = 0;

	if ( strcmp( lversion, "") == 0)
	{
	  *version = 0;
	}
	else
	{
	  sts = sscanf( lversion, "%d", version);
	  if ( sts != 1)
	    *version = 0;
	}

	strcpy( dev, ldev);
	strcpy( dir, ldir);
	strcpy( file, lfile);
	strcpy( type, ltype);
	return 1;

#elif defined (OS_LYNX) || defined(OS_LINUX)

	char	*s;
	char	ldev[200];
	char	ldir[200];
	char	lfile[200];
	char	ltype[80];

	if ( (s = strstr( filename, "::")))
	  s += 2;
	else
	  s = filename;
	strcpy( ldev, "");
	strcpy( ldir, s);

	/* Directory */
	if ( (s = strrchr( ldir, '/')))
	  s++;
	else
	  s = ldir;
	strcpy( lfile, s);
	*s = 0;

	/* File */
	if ( (s = strchr( lfile, '.')) == 0)
	  s = lfile;
	strcpy( ltype, s);
	*s = 0;

	/* Type */
	if ( (s = strchr( ltype+1, '.')) != 0)
	  *s = 0;
	if ( (s = strchr( ltype+1, ';')) != 0)
	  *s = 0;

	*version = 0;

	strcpy( dev, ldev);
	strcpy( dir, ldir);
	strcpy( file, lfile);
	strcpy( type, ltype);
	return 1;
#else
	return 0;
#endif
}  

